import logging
from datetime import datetime
from pathlib import Path

import hydra
import torch
import torch.multiprocessing as mp
from algorithm import DSFLParallelClientTrainer, DSFLServerHandler
from dataset import DSFLPartitionedDataset
from hydra.core import hydra_config
from models import DSFLModelSelector
from omegaconf import DictConfig, OmegaConf
from torch.utils.tensorboard.writer import SummaryWriter
from torchvision import transforms

from blazefl.utils import seed_everything


class DSFLPipeline:
    def __init__(
        self,
        handler: DSFLServerHandler,
        trainer: DSFLParallelClientTrainer,
        writer: SummaryWriter,
    ) -> None:
        self.handler = handler
        self.trainer = trainer
        self.writer = writer

    def main(self):
        while not self.handler.if_stop():
            round_ = self.handler.round
            # server side
            sampled_clients = self.handler.sample_clients()
            broadcast = self.handler.downlink_package()

            # client side
            self.trainer.local_process(broadcast, sampled_clients)
            uploads = self.trainer.uplink_package()

            # server side
            for pack in uploads:
                self.handler.load(pack)

            summary = self.handler.get_summary()
            for key, value in summary.items():
                self.writer.add_scalar(key, value, round_)
            logging.info(f"Round {round_}: {summary}")

        logging.info("Done!")


@hydra.main(version_base=None, config_path="config", config_name="config")
def main(
    cfg: DictConfig,
):
    print(OmegaConf.to_yaml(cfg))

    log_dir = hydra_config.HydraConfig.get().runtime.output_dir
    writer = SummaryWriter(log_dir=log_dir)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    dataset_root_dir = Path(cfg.dataset_root_dir)
    dataset_split_dir = dataset_root_dir.joinpath(timestamp)
    share_dir = Path(cfg.share_dir).joinpath(timestamp)
    state_dir = Path(cfg.state_dir).joinpath(timestamp)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    seed_everything(cfg.seed, device=device)

    dataset = DSFLPartitionedDataset(
        root=dataset_root_dir,
        path=dataset_split_dir,
        num_clients=cfg.num_clients,
        num_shards=cfg.num_shards,
        dir_alpha=cfg.dir_alpha,
        seed=cfg.seed,
        partition=cfg.partition,
        transform=transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        ),
        open_size=cfg.algorithm.open_size,
    )
    model_selector = DSFLModelSelector(num_classes=10)
    handler = DSFLServerHandler(
        model_selector=model_selector,
        model_name=cfg.model_name,
        dataset=dataset,
        global_round=cfg.global_round,
        num_clients=cfg.num_clients,
        kd_epochs=cfg.algorithm.kd_epochs,
        kd_batch_size=cfg.algorithm.kd_batch_size,
        kd_lr=cfg.algorithm.kd_lr,
        era_temperature=cfg.algorithm.era_temperature,
        open_size_per_round=cfg.algorithm.open_size_per_round,
        device=device,
        sample_ratio=cfg.sample_ratio,
    )
    trainer = DSFLParallelClientTrainer(
        model_selector=model_selector,
        model_name=cfg.model_name,
        dataset=dataset,
        share_dir=share_dir,
        state_dir=state_dir,
        seed=cfg.seed,
        device=device,
        num_clients=cfg.num_clients,
        epochs=cfg.epochs,
        batch_size=cfg.batch_size,
        lr=cfg.lr,
        kd_epochs=cfg.algorithm.kd_epochs,
        kd_batch_size=cfg.algorithm.kd_batch_size,
        kd_lr=cfg.algorithm.kd_lr,
        num_parallels=cfg.num_parallels,
    )
    pipeline = DSFLPipeline(handler=handler, trainer=trainer, writer=writer)
    try:
        pipeline.main()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt")
    except Exception as e:
        logging.exception(e)


if __name__ == "__main__":
    # NOTE: To use CUDA with multiprocessing, you must use the 'spawn' start method
    mp.set_start_method("spawn")

    main()



import multiprocessing as mp
import signal
from abc import ABC, abstractmethod
from multiprocessing.pool import ApplyResult
from pathlib import Path
from typing import Generic, TypeVar

import torch
from tqdm import tqdm

UplinkPackage = TypeVar("UplinkPackage")
DownlinkPackage = TypeVar("DownlinkPackage")


class SerialClientTrainer(ABC, Generic[UplinkPackage, DownlinkPackage]):
    """
    Abstract base class for serial client training in federated learning.

    This class defines the interface for training clients in a serial manner,
    where each client is processed one after the other.

    Raises:
        NotImplementedError: If the methods are not implemented in a subclass.
    """

    @abstractmethod
    def uplink_package(self) -> list[UplinkPackage]:
        """
        Prepare the data package to be sent from the client to the server.

        Returns:
            list[UplinkPackage]: A list of data packages prepared for uplink
            transmission.
        """
        ...

    @abstractmethod
    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None:
        """
        Process the downlink payload from the server for a list of client IDs.

        Args:
            payload (DownlinkPackage): The data package received from the server.
            cid_list (list[int]): A list of client IDs to process.

        Returns:
            None
        """
        ...


DiskSharedData = TypeVar("DiskSharedData")


class ParallelClientTrainer(
    SerialClientTrainer[UplinkPackage, DownlinkPackage],
    Generic[UplinkPackage, DownlinkPackage, DiskSharedData],
):
    """
    Abstract base class for parallel client training in federated learning.

    This class extends SerialClientTrainer to enable parallel processing of clients,
    allowing multiple clients to be trained concurrently.

    Attributes:
        num_parallels (int): Number of parallel processes to use for client training.
        share_dir (Path): Directory path for sharing data between processes.
        cache (list[UplinkPackage]): Cache to store uplink packages from clients.

    Raises:
        NotImplementedError: If the abstract methods are not implemented in a subclass.
    """

    def __init__(self, num_parallels: int, share_dir: Path, device: str) -> None:
        """
        Initialize the ParallelClientTrainer with parallelism settings.

        Args:
            num_parallels (int): Number of parallel processes to use.
            share_dir (Path): Directory path for sharing data between processes.
            device (str): Device to use for processing clients.
        """
        self.num_parallels = num_parallels
        self.share_dir = share_dir
        self.share_dir.mkdir(parents=True, exist_ok=True)
        self.device = device
        if self.device == "cuda":
            self.device_count = torch.cuda.device_count()
        self.cache: list[UplinkPackage] = []

    @abstractmethod
    def get_shared_data(self, cid: int, payload: DownlinkPackage) -> DiskSharedData:
        """
        Retrieve shared data for a given client ID and payload.

        Args:
            cid (int): Client ID.
            payload (DownlinkPackage): The data package received from the server.

        Returns:
            DiskSharedData: The shared data associated with the client ID and payload.
        """
        ...

    def get_client_device(self, cid: int) -> str:
        """
        Retrieve the device to use for processing a given client.

        Args:
            cid (int): Client ID.

        Returns:
            str: The device to use for processing the client.
        """
        if self.device == "cuda":
            return f"cuda:{cid % self.device_count}"
        return self.device

    @staticmethod
    @abstractmethod
    def process_client(path: Path, device: str) -> Path:
        """
        Process a single client based on the provided path.

        Args:
            path (Path): Path to the client's data file.
            device (str): Device to use for processing.

        Returns:
            Path: Path to the processed client's data file.
        """
        ...

    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None:
        """
        Manage the parallel processing of clients.

        This method distributes the processing of multiple clients across
        parallel processes, handling data saving, loading, and caching.

        Args:
            payload (DownlinkPackage): The data package received from the server.
            cid_list (list[int]): A list of client IDs to process.

        Returns:
            None
        """
        with mp.Pool(
            processes=self.num_parallels,
            initializer=signal.signal,
            initargs=(signal.SIGINT, signal.SIG_IGN),
        ) as pool:
            jobs: list[ApplyResult] = []
            for cid in cid_list:
                path = self.share_dir.joinpath(f"{cid}.pkl")
                data = self.get_shared_data(cid, payload)
                device = self.get_client_device(cid)
                torch.save(data, path)
                jobs.append(pool.apply_async(self.process_client, (path, device)))

            for job in tqdm(jobs, desc="Client", leave=False):
                path = job.get()
                assert isinstance(path, Path)
                package = torch.load(path, weights_only=False)
                self.cache.append(package)



from dataclasses import dataclass
from pathlib import Path

import pytest
import torch

from src.blazefl.core import ParallelClientTrainer


@dataclass
class UplinkPackage:
    cid: int
    message: str


@dataclass
class DownlinkPackage:
    message: str


@dataclass
class DiskSharedData:
    cid: int
    payload: DownlinkPackage


class DummyParallelClientTrainer(
    ParallelClientTrainer[UplinkPackage, DownlinkPackage, DiskSharedData]
):
    def uplink_package(self) -> list[UplinkPackage]:
        return self.cache

    def get_shared_data(self, cid: int, payload: DownlinkPackage) -> DiskSharedData:
        return DiskSharedData(cid=cid, payload=payload)

    @staticmethod
    def process_client(path: Path, device: str) -> Path:
        data = torch.load(path, weights_only=False)
        _ = device
        downlink_package = data.payload

        dummy_uplink_package = UplinkPackage(
            cid=data.cid, message=downlink_package.message + "<client_to_server>"
        )

        torch.save(dummy_uplink_package, path)
        return path


@pytest.mark.parametrize("num_parallels", [1, 2, 4])
@pytest.mark.parametrize("cid_list", [[], [42], [0, 1, 2]])
def test_parallel_client_trainer(
    tmp_path: Path, num_parallels: int, cid_list: list[int]
) -> None:
    trainer = DummyParallelClientTrainer(
        num_parallels=num_parallels, share_dir=tmp_path, device="cpu"
    )

    dummy_payload = DownlinkPackage(message="<server_to_client>")

    trainer.local_process(dummy_payload, cid_list)

    assert len(trainer.cache) == len(cid_list)
    for i, cid in enumerate(cid_list):
        result = trainer.cache[i]
        assert result.cid == cid
        assert result.message == "<server_to_client><client_to_server>"

    package = trainer.uplink_package()
    assert len(package) == len(cid_list)

    for i, cid in enumerate(cid_list):
        result = package[i]
        assert result.cid == cid
        assert result.message == "<server_to_client><client_to_server>"



from blazefl.core.client_trainer import ParallelClientTrainer as ParallelClientTrainer, SerialClientTrainer as SerialClientTrainer
from blazefl.core.model_selector import ModelSelector as ModelSelector
from blazefl.core.partitioned_dataset import PartitionedDataset as PartitionedDataset
from blazefl.core.server_handler import ServerHandler as ServerHandler



import torch
from _typeshed import Incomplete
from blazefl.core import ModelSelector as ModelSelector, ParallelClientTrainer as ParallelClientTrainer, PartitionedDataset as PartitionedDataset, SerialClientTrainer as SerialClientTrainer, ServerHandler as ServerHandler
from blazefl.utils import RandomState as RandomState, deserialize_model as deserialize_model, seed_everything as seed_everything, serialize_model as serialize_model
from dataclasses import dataclass
from pathlib import Path
from torch.utils.data import DataLoader

@dataclass
class FedAvgUplinkPackage:
    model_parameters: torch.Tensor
    data_size: int
    metadata: dict | None = ...

@dataclass
class FedAvgDownlinkPackage:
    model_parameters: torch.Tensor

class FedAvgServerHandler(ServerHandler):
    model: Incomplete
    dataset: Incomplete
    global_round: Incomplete
    num_clients: Incomplete
    sample_ratio: Incomplete
    device: Incomplete
    client_buffer_cache: Incomplete
    num_clients_per_round: Incomplete
    round: int
    def __init__(self, model_selector: ModelSelector, model_name: str, dataset: PartitionedDataset, global_round: int, num_clients: int, sample_ratio: float, device: str) -> None: ...
    def sample_clients(self) -> list[int]: ...
    def if_stop(self) -> bool: ...
    def load(self, payload: FedAvgUplinkPackage) -> bool: ...
    def global_update(self, buffer: list[FedAvgUplinkPackage]) -> None: ...
    @staticmethod
    def aggregate(parameters_list: list[torch.Tensor], weights_list: list[int]) -> torch.Tensor: ...
    def downlink_package(self) -> FedAvgDownlinkPackage: ...

class FedAvgSerialClientTrainer(SerialClientTrainer[FedAvgUplinkPackage, FedAvgDownlinkPackage]):
    model: Incomplete
    dataset: Incomplete
    device: Incomplete
    num_clients: Incomplete
    epochs: Incomplete
    batch_size: Incomplete
    lr: Incomplete
    optimizer: Incomplete
    criterion: Incomplete
    cache: Incomplete
    def __init__(self, model_selector: ModelSelector, model_name: str, dataset: PartitionedDataset, device: str, num_clients: int, epochs: int, batch_size: int, lr: float) -> None: ...
    def local_process(self, payload: FedAvgDownlinkPackage, cid_list: list[int]) -> None: ...
    def train(self, model_parameters: torch.Tensor, train_loader: DataLoader) -> FedAvgUplinkPackage: ...
    def evaluate(self, test_loader: DataLoader) -> tuple[float, float]: ...
    def uplink_package(self) -> list[FedAvgUplinkPackage]: ...

@dataclass
class FedAvgDiskSharedData:
    model_selector: ModelSelector
    model_name: str
    dataset: PartitionedDataset
    epochs: int
    batch_size: int
    lr: float
    cid: int
    seed: int
    payload: FedAvgDownlinkPackage
    state_path: Path

class FedAvgParallelClientTrainer(ParallelClientTrainer[FedAvgUplinkPackage, FedAvgDownlinkPackage, FedAvgDiskSharedData]):
    model_selector: Incomplete
    model_name: Incomplete
    state_dir: Incomplete
    dataset: Incomplete
    epochs: Incomplete
    batch_size: Incomplete
    lr: Incomplete
    device: Incomplete
    num_clients: Incomplete
    seed: Incomplete
    def __init__(self, model_selector: ModelSelector, model_name: str, share_dir: Path, state_dir: Path, dataset: PartitionedDataset, device: str, num_clients: int, epochs: int, batch_size: int, lr: float, seed: int, num_parallels: int) -> None: ...
    @staticmethod
    def process_client(path: Path, device: str) -> Path: ...
    @staticmethod
    def train(model: torch.nn.Module, model_parameters: torch.Tensor, train_loader: DataLoader, device: str, epochs: int, lr: float) -> FedAvgUplinkPackage: ...
    @staticmethod
    def evaulate(model: torch.nn.Module, test_loader: DataLoader, device: str) -> tuple[float, float]: ...
    def get_shared_data(self, cid: int, payload: FedAvgDownlinkPackage) -> FedAvgDiskSharedData: ...
    cache: Incomplete
    def uplink_package(self) -> list[FedAvgUplinkPackage]: ...



import random
from collections import defaultdict
from collections.abc import Mapping
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path

import torch
import torch.nn.functional as F
from dataset import DSFLPartitionedDataset
from models import DSFLModelSelector
from torch.utils.data import DataLoader, Subset

from blazefl.core import (
    ParallelClientTrainer,
    ServerHandler,
)
from blazefl.utils import (
    FilteredDataset,
    RandomState,
    seed_everything,
)


@dataclass
class DSFLUplinkPackage:
    soft_labels: torch.Tensor
    indices: torch.Tensor
    metadata: dict


@dataclass
class DSFLDownlinkPackage:
    soft_labels: torch.Tensor | None
    indices: torch.Tensor | None
    next_indices: torch.Tensor


class DSFLServerHandler(ServerHandler[DSFLUplinkPackage, DSFLDownlinkPackage]):
    def __init__(
        self,
        model_selector: DSFLModelSelector,
        model_name: str,
        dataset: DSFLPartitionedDataset,
        global_round: int,
        num_clients: int,
        sample_ratio: float,
        device: str,
        kd_epochs: int,
        kd_batch_size: int,
        kd_lr: float,
        era_temperature: float,
        open_size_per_round: int,
    ) -> None:
        self.model = model_selector.select_model(model_name)
        self.dataset = dataset
        self.global_round = global_round
        self.num_clients = num_clients
        self.sample_ratio = sample_ratio
        self.device = device
        self.kd_epochs = kd_epochs
        self.kd_batch_size = kd_batch_size
        self.kd_lr = kd_lr
        self.era_temperature = era_temperature
        self.open_size_per_round = open_size_per_round

        self.client_buffer_cache: list[DSFLUplinkPackage] = []
        self.global_soft_labels: torch.Tensor | None = None
        self.global_indices: torch.Tensor | None = None
        self.num_clients_per_round = int(self.num_clients * self.sample_ratio)
        self.round = 0

    def sample_clients(self) -> list[int]:
        sampled_clients = random.sample(
            range(self.num_clients), self.num_clients_per_round
        )

        return sorted(sampled_clients)

    def get_next_indices(self) -> torch.Tensor:
        shuffled_indices = torch.randperm(self.dataset.open_size)
        return shuffled_indices[: self.open_size_per_round]

    def if_stop(self) -> bool:
        return self.round >= self.global_round

    def load(self, payload: DSFLUplinkPackage) -> bool:
        self.client_buffer_cache.append(payload)

        if len(self.client_buffer_cache) == self.num_clients_per_round:
            self.global_update(self.client_buffer_cache)
            self.round += 1
            self.client_buffer_cache = []
            return True
        else:
            return False

    def global_update(self, buffer) -> None:
        soft_labels_list = [ele.soft_labels for ele in buffer]
        indices_list = [ele.indices for ele in buffer]
        self.metadata_list = [ele.metadata for ele in buffer]

        soft_labels_stack: defaultdict[int, list[torch.Tensor]] = defaultdict(
            list[torch.Tensor]
        )
        for soft_labels, indices in zip(soft_labels_list, indices_list, strict=True):
            for soft_label, index in zip(soft_labels, indices, strict=True):
                soft_labels_stack[int(index.item())].append(soft_label)

        global_soft_labels: list[torch.Tensor] = []
        global_indices: list[int] = []
        for indices, soft_labels in soft_labels_stack.items():
            global_indices.append(indices)
            mean_soft_labels = torch.mean(torch.stack(soft_labels), dim=0)
            # Entropy Reduction Aggregation (ERA)
            era_soft_labels = F.softmax(mean_soft_labels / self.era_temperature, dim=0)
            global_soft_labels.append(era_soft_labels)

        DSFLServerHandler.distill(
            self.model,
            self.dataset,
            global_soft_labels,
            global_indices,
            self.kd_epochs,
            self.kd_batch_size,
            self.kd_lr,
            self.device,
        )

        self.global_soft_labels = torch.stack(global_soft_labels)
        self.global_indices = torch.tensor(global_indices)

    @staticmethod
    def distill(
        model: torch.nn.Module,
        dataset: DSFLPartitionedDataset,
        global_soft_labels: list[torch.Tensor],
        global_indices: list[int],
        kd_epochs: int,
        kd_batch_size: int,
        kd_lr: float,
        device: str,
    ) -> None:
        model.to(device)
        model.train()
        openset = dataset.get_dataset(type_="open", cid=None)
        open_loader = DataLoader(
            Subset(openset, global_indices),
            batch_size=kd_batch_size,
        )
        global_soft_label_loader = DataLoader(
            FilteredDataset(
                indices=list(range(len(global_soft_labels))),
                original_data=global_soft_labels,
            ),
            batch_size=kd_batch_size,
        )
        optimizer = torch.optim.SGD(model.parameters(), lr=kd_lr)
        for _ in range(kd_epochs):
            for data, soft_label in zip(
                open_loader, global_soft_label_loader, strict=True
            ):
                data = data.to(device)
                soft_label = soft_label.to(device).squeeze(1)

                output = model(data)
                loss = F.kl_div(
                    F.log_softmax(output, dim=1), soft_label, reduction="batchmean"
                )

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
        return

    @staticmethod
    def evaulate(
        model: torch.nn.Module, test_loader: DataLoader, device: str
    ) -> tuple[float, float]:
        model.to(device)
        model.eval()
        criterion = torch.nn.CrossEntropyLoss()

        total_loss = 0.0
        total_correct = 0
        total_samples = 0
        with torch.no_grad():
            for inputs, labels in test_loader:
                inputs = inputs.to(device)
                labels = labels.to(device)

                outputs = model(inputs)
                loss = criterion(outputs, labels)

                _, predicted = torch.max(outputs, 1)
                correct = torch.sum(predicted.eq(labels)).item()

                batch_size = labels.size(0)
                total_loss += loss.item() * batch_size
                total_correct += int(correct)
                total_samples += batch_size

        avg_loss = total_loss / total_samples
        avg_acc = total_correct / total_samples

        return avg_loss, avg_acc

    def get_summary(self) -> dict[str, float]:
        server_loss, server_acc = DSFLServerHandler.evaulate(
            self.model,
            self.dataset.get_dataloader(
                type_="test",
                cid=None,
                batch_size=self.kd_batch_size,
            ),
            self.device,
        )
        client_loss = sum(m["loss"] for m in self.metadata_list) / len(
            self.metadata_list
        )
        client_acc = sum(m["acc"] for m in self.metadata_list) / len(self.metadata_list)
        return {
            "server_acc": server_acc,
            "server_loss": server_loss,
            "client_acc": client_acc,
            "client_loss": client_loss,
        }

    def downlink_package(self) -> DSFLDownlinkPackage:
        next_indices = self.get_next_indices()
        return DSFLDownlinkPackage(
            self.global_soft_labels, self.global_indices, next_indices
        )


@dataclass
class DSFLDiskSharedData:
    model_selector: DSFLModelSelector
    model_name: str
    dataset: DSFLPartitionedDataset
    epochs: int
    batch_size: int
    lr: float
    kd_epochs: int
    kd_batch_size: int
    kd_lr: float
    cid: int
    seed: int
    payload: DSFLDownlinkPackage
    state_path: Path


@dataclass
class DSFLClientState:
    random: RandomState
    model: Mapping[str, torch.Tensor]


class DSFLParallelClientTrainer(
    ParallelClientTrainer[DSFLUplinkPackage, DSFLDownlinkPackage, DSFLDiskSharedData]
):
    def __init__(
        self,
        model_selector: DSFLModelSelector,
        model_name: str,
        share_dir: Path,
        state_dir: Path,
        dataset: DSFLPartitionedDataset,
        device: str,
        num_clients: int,
        epochs: int,
        batch_size: int,
        lr: float,
        kd_epochs: int,
        kd_batch_size: int,
        kd_lr: float,
        seed: int,
        num_parallels: int,
    ) -> None:
        super().__init__(num_parallels, share_dir, device)
        self.model_selector = model_selector
        self.model_name = model_name
        self.state_dir = state_dir
        self.state_dir.mkdir(parents=True, exist_ok=True)
        self.dataset = dataset
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.kd_epochs = kd_epochs
        self.kd_batch_size = kd_batch_size
        self.kd_lr = kd_lr
        self.device = device
        self.num_clients = num_clients
        self.seed = seed

        if self.device == "cuda":
            self.device_count = torch.cuda.device_count()

    @staticmethod
    def process_client(path: Path, device: str) -> Path:
        data = torch.load(path, weights_only=False)
        assert isinstance(data, DSFLDiskSharedData)

        state: DSFLClientState | None = None
        if data.state_path.exists():
            state = torch.load(data.state_path, weights_only=False)
            assert isinstance(state, DSFLClientState)
            RandomState.set_random_state(state.random)
        else:
            seed_everything(data.seed, device=device)

        model = data.model_selector.select_model(data.model_name)

        if state is not None:
            model.load_state_dict(state.model)

        # Distill
        openset = data.dataset.get_dataset(type_="open", cid=None)
        if data.payload.indices is not None and data.payload.soft_labels is not None:
            global_soft_labels = list(torch.unbind(data.payload.soft_labels, dim=0))
            global_indices = data.payload.indices.tolist()
            DSFLServerHandler.distill(
                model=model,
                dataset=data.dataset,
                global_soft_labels=global_soft_labels,
                global_indices=global_indices,
                kd_epochs=data.kd_epochs,
                kd_batch_size=data.kd_batch_size,
                kd_lr=data.kd_lr,
                device=device,
            )

        # Train
        train_loader = data.dataset.get_dataloader(
            type_="train",
            cid=data.cid,
            batch_size=data.batch_size,
        )
        DSFLParallelClientTrainer.train(
            model=model,
            train_loader=train_loader,
            device=device,
            epochs=data.epochs,
            lr=data.lr,
        )

        # Predict
        open_loader = DataLoader(
            Subset(openset, data.payload.next_indices.tolist()),
            batch_size=data.batch_size,
        )
        soft_labels = DSFLParallelClientTrainer.predict(
            model=model,
            open_loader=open_loader,
            device=device,
        )

        # Evaluate
        val_loader = data.dataset.get_dataloader(
            type_="val",
            cid=data.cid,
            batch_size=data.batch_size,
        )
        loss, acc = DSFLServerHandler.evaulate(
            model=model,
            test_loader=val_loader,
            device=device,
        )

        package = DSFLUplinkPackage(
            soft_labels=torch.stack(soft_labels),
            indices=data.payload.next_indices,
            metadata={"loss": loss, "acc": acc},
        )

        torch.save(package, path)
        state = DSFLClientState(
            random=RandomState.get_random_state(device=device),
            model=model.state_dict(),
        )
        torch.save(state, data.state_path)
        return path

    @staticmethod
    def train(
        model: torch.nn.Module,
        train_loader: DataLoader,
        device: str,
        epochs: int,
        lr: float,
    ) -> None:
        model.to(device)
        model.train()
        optimizer = torch.optim.SGD(model.parameters(), lr=lr)
        criterion = torch.nn.CrossEntropyLoss()

        for _ in range(epochs):
            for data, target in train_loader:
                data = data.to(device)
                target = target.to(device)

                output = model(data)
                loss = criterion(output, target)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
        return

    @staticmethod
    def predict(
        model: torch.nn.Module, open_loader: DataLoader, device: str
    ) -> list[torch.Tensor]:
        model.to(device)
        model.eval()

        soft_labels = []
        with torch.no_grad():
            for data in open_loader:
                data = data.to(device)

                output = model(data)
                soft_label = F.softmax(output, dim=1)

                soft_labels.extend([sl.detach().cpu() for sl in soft_label])

        return soft_labels

    def get_shared_data(
        self, cid: int, payload: DSFLDownlinkPackage
    ) -> DSFLDiskSharedData:
        data = DSFLDiskSharedData(
            model_selector=self.model_selector,
            model_name=self.model_name,
            dataset=self.dataset,
            epochs=self.epochs,
            batch_size=self.batch_size,
            lr=self.lr,
            kd_epochs=self.kd_epochs,
            kd_batch_size=self.kd_batch_size,
            kd_lr=self.kd_lr,
            cid=cid,
            seed=self.seed,
            payload=payload,
            state_path=self.state_dir.joinpath(f"{cid}.pt"),
        )
        return data

    def uplink_package(self) -> list[DSFLUplinkPackage]:
        package = deepcopy(self.cache)
        self.cache: list[DSFLUplinkPackage] = []
        return package



import logging
from datetime import datetime
from pathlib import Path

import hydra
import torch
import torch.multiprocessing as mp
from dataset import PartitionedCIFAR10
from hydra.core import hydra_config
from models import FedAvgModelSelector
from omegaconf import DictConfig, OmegaConf
from torch.utils.tensorboard.writer import SummaryWriter
from torchvision import transforms

from blazefl.contrib import (
    FedAvgParallelClientTrainer,
    FedAvgSerialClientTrainer,
    FedAvgServerHandler,
)
from blazefl.utils import seed_everything


class FedAvgPipeline:
    def __init__(
        self,
        handler: FedAvgServerHandler,
        trainer: FedAvgSerialClientTrainer | FedAvgParallelClientTrainer,
        writer: SummaryWriter,
    ) -> None:
        self.handler = handler
        self.trainer = trainer
        self.writer = writer

    def main(self):
        while not self.handler.if_stop():
            round_ = self.handler.round
            # server side
            sampled_clients = self.handler.sample_clients()
            broadcast = self.handler.downlink_package()

            # client side
            self.trainer.local_process(broadcast, sampled_clients)
            uploads = self.trainer.uplink_package()

            metadata_list = [
                pack.metadata for pack in uploads if pack.metadata is not None
            ]
            avg_loss = sum(meta["loss"] for meta in metadata_list) / len(metadata_list)
            avg_acc = sum(meta["acc"] for meta in metadata_list) / len(metadata_list)

            logging.info(
                f"Round: {round_}, Loss: {avg_loss:.2f}, Accuracy: {avg_acc:.2f}"
            )
            self.writer.add_scalar("Loss", avg_loss, round_)
            self.writer.add_scalar("Accuracy", avg_acc, round_)

            # server side
            for pack in uploads:
                self.handler.load(pack)

        logging.info("Done!")


@hydra.main(version_base=None, config_path="config", config_name="config")
def main(
    cfg: DictConfig,
):
    print(OmegaConf.to_yaml(cfg))

    log_dir = hydra_config.HydraConfig.get().runtime.output_dir
    writer = SummaryWriter(log_dir=log_dir)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    dataset_root_dir = Path(cfg.dataset_root_dir)
    dataset_split_dir = dataset_root_dir.joinpath(timestamp)
    share_dir = Path(cfg.share_dir).joinpath(timestamp)
    state_dir = Path(cfg.state_dir).joinpath(timestamp)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    seed_everything(cfg.seed, device=device)

    dataset = PartitionedCIFAR10(
        root=dataset_root_dir,
        path=dataset_split_dir,
        num_clients=cfg.num_clients,
        num_shards=cfg.num_shards,
        dir_alpha=cfg.dir_alpha,
        seed=cfg.seed,
        partition=cfg.partition,
        transform=transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        ),
    )
    model_selector = FedAvgModelSelector(num_classes=10)
    handler = FedAvgServerHandler(
        model_selector=model_selector,
        model_name=cfg.model_name,
        dataset=dataset,
        global_round=cfg.global_round,
        num_clients=cfg.num_clients,
        device=device,
        sample_ratio=cfg.sample_ratio,
    )
    if cfg.serial:
        trainer = FedAvgSerialClientTrainer(
            model_selector=model_selector,
            model_name=cfg.model_name,
            dataset=dataset,
            device=device,
            num_clients=cfg.num_clients,
            epochs=cfg.epochs,
            lr=cfg.lr,
            batch_size=cfg.batch_size,
        )
    else:
        trainer = FedAvgParallelClientTrainer(
            model_selector=model_selector,
            model_name=cfg.model_name,
            dataset=dataset,
            share_dir=share_dir,
            state_dir=state_dir,
            seed=cfg.seed,
            device=device,
            num_clients=cfg.num_clients,
            epochs=cfg.epochs,
            lr=cfg.lr,
            batch_size=cfg.batch_size,
            num_parallels=cfg.num_parallels,
        )
    pipeline = FedAvgPipeline(handler=handler, trainer=trainer, writer=writer)
    try:
        pipeline.main()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt")
    except Exception as e:
        logging.error(e)


if __name__ == "__main__":
    # NOTE: To use CUDA with multiprocessing, you must use the 'spawn' start method
    mp.set_start_method("spawn")

    main()



"""
Federated Learning Algorithms Implementations.

This module provides implementations of various federated learning algorithms,
extending the core functionalities of BlazeFL.
"""

from blazefl.contrib.fedavg import (
    FedAvgParallelClientTrainer,
    FedAvgSerialClientTrainer,
    FedAvgServerHandler,
)

__all__ = [
    "FedAvgServerHandler",
    "FedAvgParallelClientTrainer",
    "FedAvgSerialClientTrainer",
]



from _typeshed import Incomplete
from abc import ABC, abstractmethod
from multiprocessing.pool import ApplyResult as ApplyResult
from pathlib import Path
from typing import Generic, TypeVar

UplinkPackage = TypeVar('UplinkPackage')
DownlinkPackage = TypeVar('DownlinkPackage')

class SerialClientTrainer(ABC, Generic[UplinkPackage, DownlinkPackage]):
    @abstractmethod
    def uplink_package(self) -> list[UplinkPackage]: ...
    @abstractmethod
    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None: ...
DiskSharedData = TypeVar('DiskSharedData')

class ParallelClientTrainer(SerialClientTrainer[UplinkPackage, DownlinkPackage], Generic[UplinkPackage, DownlinkPackage, DiskSharedData]):
    num_parallels: Incomplete
    share_dir: Incomplete
    device: Incomplete
    device_count: Incomplete
    cache: Incomplete
    def __init__(self, num_parallels: int, share_dir: Path, device: str) -> None: ...
    @abstractmethod
    def get_shared_data(self, cid: int, payload: DownlinkPackage) -> DiskSharedData: ...
    def get_client_device(self, cid: int) -> str: ...
    @staticmethod
    @abstractmethod
    def process_client(path: Path, device: str) -> Path: ...
    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None: ...



from blazefl.utils.dataset import FilteredDataset as FilteredDataset
from blazefl.utils.seed import RandomState as RandomState, seed_everything as seed_everything
from blazefl.utils.serialize import deserialize_model as deserialize_model, serialize_model as serialize_model



from abc import ABC, abstractmethod
from torch.utils.data import DataLoader, Dataset

class PartitionedDataset(ABC):
    @abstractmethod
    def get_dataset(self, type_: str, cid: int | None) -> Dataset: ...
    @abstractmethod
    def get_dataloader(self, type_: str, cid: int | None, batch_size: int | None) -> DataLoader: ...



from collections.abc import Callable, Sized
from pathlib import Path

import torch
import torchvision
from fedlab.utils.dataset.functional import (
    balance_split,
    client_inner_dirichlet_partition,
    shards_partition,
)
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, Dataset

from blazefl.core import PartitionedDataset
from blazefl.utils import FilteredDataset


class PartitionedCIFAR10(PartitionedDataset):
    def __init__(
        self,
        root: Path,
        path: Path,
        num_clients: int,
        seed: int,
        partition: str,
        num_shards: int | None = None,
        dir_alpha: float | None = None,
        val_ratio: float = 0.2,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
    ) -> None:
        self.root = root
        self.path = path
        self.num_clients = num_clients
        self.seed = seed
        self.partition = partition
        self.num_shards = num_shards
        self.dir_alpha = dir_alpha
        self.val_ratio = val_ratio
        self.transform = transform
        self.target_transform = target_transform

        self._preprocess()

    def _preprocess(self):
        self.root.mkdir(parents=True, exist_ok=True)
        trainset = torchvision.datasets.CIFAR10(
            root=self.root,
            train=True,
            download=True,
        )
        testset = torchvision.datasets.CIFAR10(
            root=self.root,
            train=False,
            download=True,
        )
        for type_ in ["train", "val", "test"]:
            self.path.joinpath(type_).mkdir(parents=True)

        match self.partition:
            case "client_inner_dirichlet":
                assert self.dir_alpha is not None
                client_dict = client_inner_dirichlet_partition(
                    targets=trainset.targets,
                    num_clients=self.num_clients,
                    num_classes=10,
                    dir_alpha=self.dir_alpha,
                    client_sample_nums=balance_split(
                        len(trainset.targets), self.num_clients
                    ),
                )
            case "shards":
                assert self.num_shards is not None
                client_dict = shards_partition(
                    targets=trainset.targets,
                    num_clients=self.num_clients,
                    num_shards=self.num_shards,
                )
            case _:
                raise ValueError("Invalid partition")

        for cid, indices in client_dict.items():
            train_indices, val_indices = train_test_split(
                indices,
                test_size=self.val_ratio,
            )
            client_trainset = FilteredDataset(
                train_indices,
                trainset.data,
                trainset.targets,
                transform=self.transform,
                target_transform=self.target_transform,
            )
            torch.save(client_trainset, self.path.joinpath("train", f"{cid}.pkl"))
            client_valset = FilteredDataset(
                val_indices,
                trainset.data,
                trainset.targets,
                transform=self.transform,
                target_transform=self.target_transform,
            )
            torch.save(client_valset, self.path.joinpath("val", f"{cid}.pkl"))

        torch.save(
            testset,
            self.path.joinpath("test", "test.pkl"),
        )

    def get_dataset(self, type_: str, cid: int | None) -> Dataset:
        match type_:
            case "train" | "val":
                dataset = torch.load(
                    self.path.joinpath(type_, f"{cid}.pkl"),
                    weights_only=False,
                )
            case "test":
                dataset = torch.load(
                    self.path.joinpath(type_, "test.pkl"), weights_only=False
                )
            case _:
                raise ValueError("Invalid type_")
        assert isinstance(dataset, Dataset)
        return dataset

    def get_dataloader(
        self, type_: str, cid: int | None, batch_size: int | None = None
    ) -> DataLoader:
        dataset = self.get_dataset(type_, cid)
        assert isinstance(dataset, Sized)
        batch_size = len(dataset) if batch_size is None else batch_size
        data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
        return data_loader
